"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var schema_1 = require("./schema");
var ast_parser_1 = require("../parser/ast-parser");
var index_1 = require("../data/index");
var io_error_codes_1 = require("../errors/io-error-codes");
var constants_1 = require("../parser/constants");
var io_error_1 = require("../errors/io-error");
var is_1 = require("../utils/is");
var KeyValueCollection = /** @class */ (function () {
    function KeyValueCollection() {
        this._keys = [];
        this._map = {};
    }
    KeyValueCollection.prototype.length = function () {
        return this._keys.length;
    };
    Object.defineProperty(KeyValueCollection.prototype, "keys", {
        get: function () {
            return __spreadArrays(this._keys);
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Gets the varialbe value. This function is intend to be used internally
     * for quickly fetching the variable value, hence it accepts any key to keep the
     * consumer code free from type checking. The function validate the key and
     * returns the associated value, if avialable. Otherwise returns undefined.
     * @param key {any} The varialbe key starting with $
     * @returns The value associated with the variable
     */
    KeyValueCollection.prototype.getV = function (key) {
        if (is_1.isString(key) && key.startsWith('$') && key.length > 1) {
            return this.get(key.substr(1));
        }
    };
    /**
     * Gets a value for specified key
     * @param key {string} The key
     * @returns Value
     */
    KeyValueCollection.prototype.get = function (key) {
        return this._map[key];
    };
    /**
     * Sets the value for the key.
     * @param key {string} The key
     * @param val {any} The associated value for that key.
     */
    KeyValueCollection.prototype.set = function (key, val) {
        if (key in this._map === false) {
            this._keys.push(key);
        }
        this._map[key] = val;
        return this;
    };
    /**
     * Removes the specified key from the collection.
     * @param key {string} The key of the collection item, which needs to be removed.
     */
    KeyValueCollection.prototype.remove = function (key) {
        if (key in this._map === false)
            return this;
        var index = this._keys.indexOf(key);
        this._keys.splice(index);
        delete this._map[key];
        return this;
    };
    Object.defineProperty(KeyValueCollection.prototype, "schema", {
        /**
         * Gets schema
         */
        get: function () {
            return this._map[constants_1.SCHEMA] || null;
        },
        /**
         * Sets schema
         */
        set: function (schema) {
            if (!this.schema) {
                this._keys.push(constants_1.SCHEMA);
            }
            this._map[constants_1.SCHEMA] = schema;
        },
        enumerable: false,
        configurable: true
    });
    /**
     *
     * @param collection The collection string that needs to be compiled!
     * @param schema The schema
     */
    KeyValueCollection.compile = function (collection, schema) {
        var tree;
        var newCollection = new KeyValueCollection();
        if (is_1.isString(collection)) {
            var parser = new ast_parser_1.default(collection, true);
            parser.parse();
            tree = parser.header;
        }
        else if (is_1.isParserTree(collection)) {
            tree = collection;
        }
        else {
            throw new Error('invlid-collection');
        }
        // If it is object, it must be schema. Then convert it into
        // collection.
        if (tree.type === 'object') {
            var compiledSchema = schema_1.default.compile(tree);
            newCollection.set(constants_1.SCHEMA, compiledSchema);
        }
        // If it not collection, throw and invalid header error
        else if (tree.type === 'collection') {
            // Setup new header
            _parseCollection(tree, newCollection, function (key, val) {
                newCollection._keys.push(key);
                newCollection._map[key] = val;
            });
        }
        else {
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invlidHeader, 'Invalid value found in header');
        }
        // Override the schema with the supplied one
        if (schema) {
            var compiledSchema = schema;
            if (is_1.isString(schema)) {
                compiledSchema = schema_1.default.compile(tree);
            }
            newCollection.set(constants_1.SCHEMA, compiledSchema);
        }
        return newCollection;
    };
    return KeyValueCollection;
}());
exports.default = KeyValueCollection;
function _parseCollection(tree, collection, callback) {
    var map = {};
    var keys = [];
    for (var index = 0; index < tree.values.length; index += 1) {
        var item = tree.values[index];
        // Verify item is an object and contains only 1 key-value pair!
        if (!is_1.isParserTree(item)) {
            console.warn('>>', tree);
            // TODO: Throw better error
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidHeaderItem);
        }
        if (item.type !== 'object') {
            // TODO: Throw better error
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidHeaderItem);
        }
        if (item.values.length !== 1) {
            console.log('***', item, item.values);
            // TODO: Throw better error
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidHeaderItem);
        }
        var keyVal = item.values[0];
        if (!is_1.isKeyVal(keyVal)) {
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidHeaderItem);
        }
        var key = keyVal.key;
        var value = keyVal.value;
        // When key is a SCHEMA, compile the value and create schema
        if (key.startsWith('$')) {
            var val = schema_1.default.compile(value, collection);
            callback(key, val);
        }
        else if (is_1.isParserTree(value)) {
            var val = index_1.default.parse(value, collection);
            callback(key, val);
        }
        else if (is_1.isToken(value)) {
            var val = value.value;
            // If the token represents a variable in collection, replace
            // set the key with the variable value.
            if (value.type === 'string' && val.length > 2 && val[0] === '$') {
                var varVal = collection.get(value.value.substr(1));
                if (varVal !== undefined) {
                    val = varVal;
                }
            }
            callback(key, val);
        }
        // When value is null or of KeyVal type!
        else {
            // TODO: Fix this error
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidHeaderItem);
        }
    }
}
//# sourceMappingURL=index.js.map