"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("../types/index");
var index_1 = require("../data/index");
var ast_parser_1 = require("../parser/ast-parser");
var io_error_codes_1 = require("../errors/io-error-codes");
var io_error_1 = require("../errors/io-error");
var typedef_registry_1 = require("../types/typedef-registry");
var is_1 = require("../utils/is");
/**
 * Represents the Internet Object Schema class which is responsible for
 * compiling schemas and applying them to data objects.
 */
var Schema = /** @class */ (function () {
    /**
     *
     * @param schema Initializes the new instance of the `Schema` class.
     *
     * @param schema {string, ASTParser}
     */
    function Schema(schema) {
        this._schema = schema;
    }
    Object.defineProperty(Schema.prototype, "schema", {
        /**
         * Returns the underlaying schema structure
         */
        get: function () {
            return this._schema;
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Applies the schema to specified data and returns the
     * mapped
     */
    Schema.prototype.apply = function (data, vars) {
        if (data === undefined || data === null) {
            return data;
        }
        return _apply(data, this._schema, {}, vars);
    };
    /**
     * Converts and returns the string version of current schema.
     */
    Schema.prototype.toString = function () {
        // TODO: Need to work on this function.
    };
    Schema.compile = function (schema, vars) {
        var parsedSchema = null;
        // When string is passed, first parse it with ASTParser
        if (is_1.isString(schema)) {
            var parser = new ast_parser_1.default(schema, true);
            // TODO: Ensure what happens when parsing is failed
            parser.parse();
            parsedSchema = parser.header;
        }
        else if (is_1.isParserTree(schema)) {
            parsedSchema = schema;
        }
        if (is_1.isToken(parsedSchema)) {
            throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.invalidSchema, 'Expecting a valid schema', parsedSchema);
        }
        var compiledSchema = _compileObjectSchema(parsedSchema, undefined, vars);
        return new Schema(compiledSchema);
    };
    return Schema;
}());
exports.default = Schema;
var _getCompileValue = function (value, path, vars) {
    // Process Token
    if (is_1.isToken(value)) {
        if (vars && is_1.isString(value.value)) {
            var foundValue = vars.getV(value.value);
            // In the case of variable, if value is found, return
            if (foundValue !== undefined) {
                if (foundValue instanceof Schema) {
                    return {
                        type: 'object',
                        schema: foundValue.schema,
                        path: path
                    };
                }
                return foundValue;
            }
        }
        if (is_1.isDataType(value.value)) {
            return {
                type: value.value,
                path: path
            };
        }
        // return value.value
        // console.log("+++", value.value)
        // TODO: Throw better  error
        throw new io_error_1.InternetObjectError('invalid-datatype', 'Invalid data type', value);
    }
    if (is_1.isParserTree(value)) {
        if (_isMemberDef(value)) {
            return _compileMemberDefTree(value, path, vars);
        }
        var arrayDef = value.type === 'array';
        return {
            type: value.type,
            schema: arrayDef
                ? _compileArraySchema(value, path, vars)
                : _compileObjectSchema(value, path, vars)
        };
    }
    // TODO: Throw better error
    throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidValue);
};
var _compileMemberDefTree = function (root, path, vars) {
    if (root.values.length === 0) {
        return {
            type: root.type,
            schema: root.type === 'object' ? {} : [],
            path: path
        };
    }
    var firstVal = root.values[0];
    // Array Root
    if (root.type === 'array') {
        // Arrays can't have multiple schemas
        if (root.values.length > 1) {
            // TODO: Throw better  error
            throw new io_error_1.InternetObjectError('invalid-array-definition');
        }
        if (is_1.isToken(firstVal)) {
            if (is_1.isDataType(firstVal.value)) {
                return {
                    type: 'array',
                    schema: {
                        type: firstVal.value,
                        path: _concatPath('[', path)
                    },
                    path: path
                };
            }
            else {
                // TODO: Throw better  error
                throw new io_error_1.InternetObjectError('invalid-array-definition');
            }
        }
        else if (is_1.isParserTree(firstVal)) {
            return {
                type: 'array',
                schema: _compileArraySchema(firstVal, path, vars),
                path: path
            };
        }
    }
    // Object Root
    var memberDef = {};
    if (is_1.isToken(firstVal) && is_1.isString(firstVal.value)) {
        var ioType = _varOrVal(firstVal, vars);
        if (is_1.isDataType(ioType)) {
            memberDef.type = ioType;
        }
        else {
            throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.invalidType, ioType, firstVal);
        }
    }
    else if (is_1.isParserTree(firstVal)) {
        memberDef.type = firstVal.type;
        memberDef.schema =
            firstVal.type === 'object'
                ? _compileObjectSchema(firstVal, path, vars)
                : _compileArraySchema(firstVal, path, vars);
    }
    for (var index = 1; index < root.values.length; index++) {
        var item = root.values[index];
        if (is_1.isKeyVal(item)) {
            var val = item.value;
            if (is_1.isToken(val)) {
                memberDef[item.key] = _varOrVal(val, vars);
            }
            else if (is_1.isParserTree(val)) {
                memberDef[item.key] = index_1.default.parse(val, vars);
            }
            else {
                // TODO: Consider this case when memberdef key = {object value}
                console.warn('Check this case', 'invalid-option', val);
                // throw new InternetObjectError("invalid-option")
            }
        }
        else if (is_1.isToken(item)) {
            memberDef[item.value] = true;
        }
        else {
            // TODO: Throw better error
            throw new io_error_1.InternetObjectError('invalid-value');
        }
    }
    memberDef.path = path;
    return memberDef;
};
var _varOrVal = function (token, vars) {
    if (is_1.isString(token.value) && vars) {
        var val = vars.getV(token.value);
        return val !== undefined ? val : token.value;
    }
    return token.value;
};
var _compileArraySchema = function (root, path, vars) {
    var array = root.values;
    var arrayPath = _concatPath('[', path);
    if (array.length === 0) {
        return {
            type: 'any',
            path: _concatPath('[', path)
        };
    }
    if (_isMemberDef(root)) {
        return _compileMemberDefTree(root, _concatPath('[', path), vars);
    }
    else if (is_1.isParserTree(root)) {
        return {
            type: root.type,
            schema: root.type === 'object'
                ? _compileObjectSchema(root, arrayPath, vars)
                : _compileArraySchema(root, arrayPath, vars),
            path: arrayPath
        };
    }
    // TODO: Throw better error
    throw new io_error_1.InternetObjectError('invalid-array-schema');
};
var _compileObjectSchema = function (root, path, vars) {
    if (root.values.length === 0)
        return {};
    var schema = {
        keys: [],
        defs: {}
    };
    for (var index = 0; index < root.values.length; index += 1) {
        var item = root.values[index];
        // Item = Tree
        if (is_1.isParserTree(item)) {
            // TODO: Throw better error
            throw new io_error_1.InternetObjectError('key-required', 'Encountered an array or an object while expecting key.');
        }
        // Item = KeyVal
        else if (is_1.isKeyVal(item)) {
            var _a = _parseKey(item.key), name_1 = _a.name, optional = _a.optional, nullable = _a.nullable;
            var currentPath = _concatPath(name_1, path);
            var value = _getCompileValue(item.value, currentPath, vars);
            schema.keys.push(name_1);
            // When value is a memberDef
            if (value.type) {
                value.optional = value.optional || optional || false;
                value.null = value.null || nullable || false;
                value.path = currentPath;
                schema.defs[name_1] = value;
            }
            else if (is_1.isDataType(value)) {
                schema.defs[name_1] = {
                    type: value,
                    optional: optional || false,
                    null: nullable || false,
                    path: currentPath
                };
            }
            else {
                throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.invalidType, JSON.stringify(value), item);
            }
        }
        else if (is_1.isToken(item)) {
            var value = _varOrVal(item, vars);
            var _b = _parseKey(value), name_2 = _b.name, optional = _b.optional, nullable = _b.nullable;
            var currentPath = _concatPath(name_2, path);
            schema.keys.push(name_2);
            schema.defs[name_2] = {
                type: 'any',
                optional: optional || false,
                null: nullable || false,
                path: currentPath
            };
        }
    }
    return schema;
};
var _isMemberDef = function (root) {
    if (!root || !root.values)
        return false;
    if (root.values.length === 0)
        return false;
    var first = root.values[0];
    if (is_1.isToken(first) && is_1.isDataType(first.value)) {
        return true;
    }
    else if (is_1.isParserTree(first)) {
        return true;
    }
    return false;
};
var _parseKey = function (key) {
    var optionalExp = /\?$/;
    var nullExp = /\*$/;
    var optNullExp = /\?\*$/;
    if (!is_1.isString(key)) {
        throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.invalidKey);
    }
    if (key.match(optNullExp)) {
        return {
            name: key.substr(0, key.length - 2),
            optional: true,
            nullable: true
        };
    }
    else if (key.match(nullExp)) {
        return {
            name: key.substr(0, key.length - 1),
            optional: false,
            nullable: true
        };
    }
    else if (key.match(optionalExp) !== null) {
        return {
            name: key.substr(0, key.length - 1),
            optional: true,
            nullable: false
        };
    }
    return { name: key, optional: false, nullable: false };
};
// Concats the path
var _concatPath = function (newPath, oldPath) {
    var path = newPath;
    if (oldPath) {
        path =
            oldPath.endsWith('[') || newPath === '[' ? "" + oldPath + newPath : oldPath + "." + newPath;
    }
    return path;
};
function _apply(data, schema, container, vars) {
    var objectDef = typedef_registry_1.TypedefRegistry.get('object');
    var schemaDef = {
        type: 'object',
        path: '',
        schema: schema
    };
    if (is_1.isParserTree(data)) {
        if (data.type === 'collection') {
            var collection_1 = [];
            data.values.forEach(function (item) {
                collection_1.push(objectDef.parse(item, schemaDef, vars));
            });
            return collection_1;
        }
        return objectDef.parse(data, schemaDef, vars);
    }
    else if (is_1.isArray(data)) {
        var collection = data.forEach(function (item) {
            return objectDef.load(item, schemaDef);
        });
        return collection;
    }
    else {
        return objectDef.load(data, schemaDef);
    }
}
//# sourceMappingURL=schema.js.map