"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.InternetObject = void 0;
var index_1 = require("./header/index");
var index_2 = require("./data/index");
var schema_1 = require("./header/schema");
var ast_parser_1 = require("./parser/ast-parser");
var io_error_codes_1 = require("./errors/io-error-codes");
var is_1 = require("./utils/is");
var constants_1 = require("./parser/constants");
var io_error_1 = require("./errors/io-error");
var index_3 = require("./serializer/index");
/**
 * Represents the InternetObject. A main facade class that
 * takes care of parsing, schema processing and other
 * core responsibilities.
 *
 * @public
 */
var InternetObject = /** @class */ (function () {
    /**
     * Initializes the new instance of the `InternetObject`
     * @param o {string, object} An internet object text or POJO which needs to be parsed or loaded
     * @param schema {string, Schema} An optional `Schema` object or its `string` representation
     */
    function InternetObject(o, schema) {
        var header;
        var data;
        if (is_1.isString(o)) {
            var parsed = _parse(o, schema);
            header = parsed.header;
            data = parsed.data;
        }
        else {
            var parsed = _parseObject(o, schema);
            data = parsed.data;
            header = parsed.header;
        }
        this._data = data;
        this._header = header;
    }
    /**
     * Create or define definitions.
     * @param callSite
     * @param substitutions
     */
    InternetObject.doc = function (callSite) {
        if (callSite === void 0) { callSite = {}; }
        var substitutions = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            substitutions[_i - 1] = arguments[_i];
        }
        var text = _getText(callSite, substitutions);
        return new InternetObject(text);
    };
    // function pro
    /**
     * Create or define definitions.
     * @param callSite
     * @param substitutions
     */
    InternetObject.defs = function (callSite) {
        if (callSite === void 0) { callSite = {}; }
        var substitutions = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            substitutions[_i - 1] = arguments[_i];
        }
        var text = _getText(callSite, substitutions);
        return index_1.default.compile(text);
    };
    Object.defineProperty(InternetObject.prototype, "header", {
        /**
         * Gets the header section of the object.
         */
        get: function () {
            return this._header;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(InternetObject.prototype, "schema", {
        /**
         * Gets the schema associated with the object.
         */
        get: function () {
            return this._header.schema;
        },
        enumerable: false,
        configurable: true
    });
    InternetObject.prototype.serialize = function () {
        return index_3.default.serialize(this);
    };
    Object.defineProperty(InternetObject.prototype, "data", {
        /**
         * Gets the data section of the object
         */
        get: function () {
            return this._data;
        },
        enumerable: false,
        configurable: true
    });
    return InternetObject;
}());
exports.InternetObject = InternetObject;
function _getText(callSite) {
    if (callSite === void 0) { callSite = {}; }
    var substitutions = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        substitutions[_i - 1] = arguments[_i];
    }
    var template;
    try {
        template = Array.from(callSite.raw);
    }
    catch (e) {
        throw new TypeError('Cannot convert undefined or null to object');
    }
    var text = template
        .map(function (chunk, i) {
        if (callSite.raw.length <= i) {
            return chunk;
        }
        return substitutions[i - 1] !== undefined ? substitutions[i - 1] + chunk : chunk;
    })
        .join('');
    return text;
}
function _getCompiledSchema(schema) {
    if (!schema)
        return null;
    if (schema instanceof schema_1.default) {
        return schema;
    }
    return schema_1.default.compile(schema);
}
function _parseObject(o, schema) {
    var compiledSchema = _getCompiledSchema(schema);
    if (compiledSchema === null) {
        return {
            data: o,
            header: new index_1.default()
        };
    }
    return {
        data: compiledSchema.apply(o),
        header: new index_1.default().set(constants_1.SCHEMA, compiledSchema)
    };
}
// Parses the Internet Object string and returns the value
function _parse(text, schema) {
    var parser = new ast_parser_1.default(text);
    var compiledSchema = null;
    var compiledHeader = null;
    // Parse the text
    parser.parse();
    if (schema) {
        if (is_1.isString(schema)) {
            compiledSchema = schema_1.default.compile(schema);
        }
        else if (schema instanceof schema_1.default) {
            compiledSchema = schema;
        }
        else {
            // TODO: Throw better error
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidSchema);
        }
    }
    var header;
    if (parser.header) {
        // console.warn(">>>", compiledSchema, parser.header)
        if (compiledSchema === null) {
            header = index_1.default.compile(parser.header);
        }
        else {
            header = index_1.default.compile(parser.header, compiledSchema);
        }
        if (header.schema) {
            compiledSchema = header.schema;
        }
    }
    else {
        header =
            compiledSchema === null
                ? new index_1.default()
                : new index_1.default().set(constants_1.SCHEMA, compiledSchema);
    }
    var data = compiledSchema
        ? compiledSchema.apply(parser.data, header)
        : index_2.default.parse(parser.data, header);
    return { header: header, data: data };
}
//# sourceMappingURL=internet-object.js.map