"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tokenizer_1 = require("./tokenizer");
var io_error_codes_1 = require("../errors/io-error-codes");
var io_error_1 = require("../errors/io-error");
var is_1 = require("../utils/is");
var constants_1 = require("./constants");
var NOT_STARTED = 'not-started';
var STARTED = 'started';
var FINISHED = 'finished';
var ERROR = 'error';
/**
 * Represents the which accepts the raw text and generates abstract syntax tree.
 * The format of the generated AST is somewhat similar to the standard one,
 * but not the same.
 *
 * @internal
 */
var ASTParser = /** @class */ (function () {
    /**
     * Initializes the new instance of `ASTParser` class.
     * @param text {string} The text which needs to be parsed
     * @param isHeader {boolean} Whether the text is schema or not, `true` if it is schema.
     */
    function ASTParser(text, isHeader) {
        var _this = this;
        if (isHeader === void 0) { isHeader = false; }
        this._isCollection = false;
        this._isHeaderOpen = true;
        this._lastToken = undefined;
        this._stack = [];
        this._tree = {};
        this._isFinalToken = false;
        // TODO: Don't use public fields, use btter appraoch
        // such as readonly properties like parser.isFinished...
        this.status = NOT_STARTED;
        /**
         * Parses the text.
         */
        this.parse = function () {
            var tokenizer = _this._tokenizer;
            while (!tokenizer.done) {
                var token = tokenizer.read();
                _this._isFinalToken = tokenizer.done;
                // When the blank string is passed to the parser, tokenizer
                // returns the null value
                if (token !== null) {
                    _this._process(token);
                }
            }
            _this._finalize();
        };
        // Process the tokens and build ast.
        this._process = function (token) {
            var isFinalToken = _this._isFinalToken;
            var stack = _this._stack;
            // Do not process comments!
            if (token.type === 'comment') {
                return;
            }
            _this.status = STARTED;
            if (token.type === 'sep' || token.type === 'datasep') {
                // Push new object
                if (token.value === '{') {
                    _this._push('object', token.index, token.row, token.col);
                    // console.log("#####", this._stack)
                }
                // Push new array
                else if (token.value === '[') {
                    _this._push('array', token.index, token.row, token.col);
                }
                // When a colon is found, consider previous value as key
                // and keep the value slot to null. When the next token arives, push it
                // into the value slot.
                // TODO: Todo check key-value pair is complete.
                else if (token.value === constants_1.COLON) {
                    var obj = _this._getObject();
                    // TODO: Validate length of the obj.values
                    var lastToken = obj.values[obj.values.length - 1];
                    if (is_1.isToken(lastToken) &&
                        ['string', 'number', 'boolean'].indexOf(typeof lastToken.value) > -1) {
                        obj.values[obj.values.length - 1] = {
                            key: lastToken.value,
                            value: undefined,
                            index: lastToken.index,
                            row: lastToken.row,
                            col: lastToken.col
                        };
                    }
                    else {
                        // TODO:Handle null and other types
                        // Throw Invalid key error!
                    }
                }
                // New Token
                else if (token.value === constants_1.COMMA) {
                    // Handle consicutive ending commas such as `test,,`
                    if (isFinalToken &&
                        _this._lastToken !== undefined &&
                        (_this._lastToken.value === constants_1.COMMA || _this._lastToken.value === constants_1.TILDE)) {
                        _this._addValue(undefined, _this._lastToken.index, _this._lastToken.row, _this._lastToken.col);
                        _this._addValue(undefined, token.index, token.row, token.col);
                    }
                    else if (
                    // When the last token is comma such as `value,`
                    isFinalToken ||
                        // When the first token is comma such as `,value`
                        _this._lastToken === undefined ||
                        // When comma is found after opening of braces, colon or datasep operators
                        [constants_1.CURLY_OPEN, constants_1.SQUARE_OPEN, constants_1.COLON, constants_1.DATASEP].indexOf(_this._lastToken.value) > -1 ||
                        // When the comma is found just after another comma
                        _this._lastToken.value === constants_1.COMMA ||
                        // When the comma is found just after the tilde (collection) sign
                        _this._lastToken.value === constants_1.TILDE) {
                        _this._addValue(undefined, token.index, token.row, token.col);
                    }
                }
                // Close object and close array
                else if (token.value === constants_1.CURLY_CLOSED || token.value === constants_1.SQUARE_CLOSED) {
                    _this._pop(token.value === constants_1.CURLY_CLOSED ? 'object' : 'array', token);
                }
                else if (token.value === constants_1.TILDE) {
                    // The ~ sign can occur in any section (header/data) first.
                    // If it is a new tree, or header/schema is already been established,
                    // set the collection mode to true.
                    if (_this._stack.length === 0) {
                        _this._isCollection = true;
                    }
                    // When the TILDE is encountered second time check collection mode is on,
                    // if not, throw an error
                    else if (!_this._isCollection && !_this._tree.data) {
                        // TODO: Throw an error here
                        throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.invalidCollection, undefined, token);
                    }
                    else {
                        _this._processObject();
                    }
                }
                // Ready for schema generation
                else if (token.value === constants_1.DATASEP) {
                    // Header is closed when, datasep is found. After the header
                    // is closed, throw an error when another datasep is sent!
                    if (!_this._isHeaderOpen) {
                        throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.multipleHeaders, 'Multiple headers are not allowed.', token);
                    }
                    _this._processObject();
                    _this._isHeaderOpen = false;
                    _this._isCollection = false;
                    _this._stack.length = 0;
                    // this._addToHeader()
                }
            }
            else {
                _this._addValue(token, token.index, token.row, token.col);
            }
            _this._lastToken = token;
        };
        this._processObject = function () {
            var isCollection = _this._isCollection;
            var isHeaderOpen = _this._isHeaderOpen;
            var data = _this._stack[0];
            // Nothing to process
            if (_this._stack.length === 0) {
                // TODO: Handle this case
                return;
            }
            // When there are open objects and arrays, stack will contain
            // more than one items.
            if (_this._stack.length > 1) {
                // TODO: Throw a proper error here
                throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.openBracket, 'Expecting bracket to be closed', _this._lastToken || undefined);
            }
            // Process header section
            if (isHeaderOpen) {
                if (isCollection) {
                    if (!_this._tree.header) {
                        _this._tree.header = {
                            type: 'collection',
                            values: [data]
                        };
                    }
                    else {
                        _this._tree.header.values.push(data);
                    }
                }
                else {
                    _this._tree.header = data;
                }
            }
            else {
                // Process data section
                if (isCollection) {
                    if (!_this._tree.data) {
                        _this._tree.data = {
                            type: 'collection',
                            values: [data]
                        };
                    }
                    else {
                        // console.log("3.", data)
                        _this._tree.data.values.push(data);
                    }
                }
                else {
                    _this._tree.data = data;
                }
            }
            // Reset stack
            _this._stack.length = 0;
        };
        this._finalize = function () {
            _this.status = FINISHED;
            _this._processObject();
            // console.warn(JSON.stringify(this._tree, null, 2))
            // TODO: Consider the scenario when more than 1 item exists in the stack!
            if (_this._stack.length > 1) {
                console.warn('Invalid Stack', _this._stack);
            }
            if (_this._isHeaderOpen) {
                _this._tree.data = _this._tree.header;
                delete _this._tree.header;
                if (_this._tree.data) {
                    if (_this.data.values.length === 1 &&
                        _this._isCollection === false &&
                        is_1.isToken(_this.data.values[0])) {
                        _this._tree.data.type = 'scalar';
                    }
                    // TODO: Add comment
                    else if (_this._tree.data.type === 'object' && _this._tree.data.values.length === 1) {
                        _this._tree.data = _this._tree.data.values[0];
                    }
                }
            }
            if (_this._isHeader) {
                _this._tree.header = _this._tree.data;
                _this._tree.data = undefined; // TODO: Replace this with `undefined`
            }
        };
        this._getOrCreateObject = function (index, row, col) {
            if (_this._stack.length === 0) {
                _this._stack.push({
                    type: 'object',
                    values: [],
                    index: index,
                    row: row,
                    col: col
                });
            }
            var obj = _this._stack[_this._stack.length - 1];
            return obj;
        };
        this._getObject = function () {
            // TODO: Verify and remove the commented code!
            // if (this._stack.length === 0) {
            //   this._stack.push({
            //     type: 'object',
            //     values: []
            //   })
            // }
            var obj = _this._stack[_this._stack.length - 1];
            return obj;
        };
        this._addValue = function (value, index, row, col) {
            var obj = _this._getOrCreateObject(index, row, col);
            // If last token is : then
            if (_this._lastToken && _this._lastToken.value === ':') {
                var lastVal = obj.values[obj.values.length - 1];
                // When the lastVal is a KeyVal, set the current token as
                // the value of lastVal
                if (lastVal && is_1.isKeyVal(lastVal)) {
                    lastVal.value = value;
                }
                else {
                    // TODO: Verify this case!
                    console.warn('Verify this case!');
                }
            }
            else {
                _this._checkValueSlot(value);
                obj.values.push(value);
            }
        };
        this._push = function (type, index, row, col, values) {
            if (type === void 0) { type = 'object'; }
            if (values === void 0) { values = []; }
            var object = _this._getOrCreateObject(index, row, col);
            var newObject = { type: type, values: values, index: index, row: row, col: col };
            _this._addValue(newObject, index, row, col);
            _this._stack.push(newObject);
        };
        this._pop = function (type, token) {
            var obj = _this._getObject();
            if (obj.type !== type) {
                var message = "Expecting \"" + (obj.type === 'object' ? '}' : ']') + "\" but found \"" + (obj.type === 'object' ? ']' : '}') + "\"";
                _this.status = ERROR;
                throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.invalidBracket, message, token);
            }
            _this._stack.pop();
        };
        this._text = text;
        this._tokenizer = new tokenizer_1.default(text);
        this._isHeader = isHeader;
    }
    Object.defineProperty(ASTParser.prototype, "header", {
        /**
         * Returns the header part. Throws an error if invoked
         * before the parsing is finished!
         */
        get: function () {
            if (this.status !== FINISHED) {
                throw new Error('parsing-not-finished');
            }
            return this._tree.header;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(ASTParser.prototype, "data", {
        /**
         * Returns the data part. Throws an error if invoked
         * before the parsing is finished!
         */
        get: function () {
            if (this.status !== FINISHED) {
                throw new Error('parsing-not-finished');
            }
            return this._tree.data;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(ASTParser.prototype, "text", {
        /**
         * Returns the text this object has parsed or about
         * to be parsed.
         */
        get: function () {
            return this._text;
        },
        enumerable: false,
        configurable: true
    });
    // Ensures values can only be added after a separator
    ASTParser.prototype._checkValueSlot = function (value) {
        var token = this._lastToken;
        if (token === undefined)
            return;
        if (value === undefined)
            return; // TODO: Check this, replaced null with undefined
        if (is_1.isToken(value) && value.value === '')
            return;
        if ([',', ':', '~', '{', '[', '---'].indexOf(token.value) === -1) {
            // console.warn(token, value)
            // TODO: Provide better error
            throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.expectingSeparator, '', value);
        }
    };
    return ASTParser;
}());
exports.default = ASTParser;
//# sourceMappingURL=ast-parser.js.map