"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var constants_1 = require("./constants");
var io_error_codes_1 = require("../errors/io-error-codes");
var io_error_1 = require("../errors/io-error");
var is_1 = require("../utils/is");
var escapeCharMap = {
    '\\': '\\',
    '/': '/',
    '"': '"',
    b: '\b',
    f: '\f',
    n: '\n',
    r: '\r',
    t: '\t'
};
var Tokenizer = /** @class */ (function () {
    function Tokenizer(text) {
        var _this = this;
        this._tokens = [];
        // Status
        this._value = '';
        this._lastToken = null;
        this._tokenLength = 0;
        this._done = false;
        // Positions
        this._index = -1;
        this._row = 1;
        this._col = 0;
        this._start = -1;
        this._end = -1;
        // Flags
        this._isQuoatedString = false;
        this._isRawString = false;
        this._isEscaping = false;
        this._isCommenting = false;
        this.read = function () {
            if (_this.done)
                return null;
            var text = _this._text;
            _this._lastToken = {};
            _this._value = '';
            _this._start = -1;
            _this._end = -1;
            _this._tokenLength = 0;
            _this._isQuoatedString = false;
            var token = _this._next();
            // No character has found while reading the token yet
            if (_this._done && token !== null && is_1.isUndefined(token.col)) {
                return null;
            }
            return token;
        };
        this.readAll = function () {
            // let token = this.read()
            while (!_this._done) {
                _this.read();
            }
            return _this;
        };
        this.push = function () {
            var _a;
            var items = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                items[_i] = arguments[_i];
            }
            (_a = _this._tokens).push.apply(_a, items);
            return _this;
        };
        this._returnToken = function () {
            var token = _this._lastToken;
            if (token === null)
                return null;
            // if (this._start === -1 || this._end === -1) return null
            var type = 'string';
            var value = _this._value;
            token.token = _this._text.substring(_this._start, _this._end + (_this._isQuoatedString ? 1 : 1));
            // console.log(JSON.stringify(token.token, null, 2), value)
            // const confirmedString = token.token.endsWith(STRING_ENCLOSER) || token.token.endsWith(QUOTE)
            // Trim the white spaces only when the strings does not ends with
            // the string encloser.
            // if (!confirmedString) {
            //   value = this._value.trim()
            // }
            // Process quoted string
            if (token.token.startsWith(constants_1.STRING_ENCLOSER)) {
                if (!token.token.endsWith(constants_1.STRING_ENCLOSER)) {
                    throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.stringNotClosed, 'String not closed, expecting " at the end of the string.', token);
                }
            }
            // Process raw string
            else if (token.token.startsWith(constants_1.QUOTE)) {
                if (!token.token.endsWith(constants_1.QUOTE)) {
                    throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.stringNotClosed, "String not closed, expecting ' at the end of the string.", token);
                }
            }
            else if (token.token.startsWith(constants_1.HASH)) {
                type = 'comment';
                value = value.substr(1).trim();
            }
            // Process open string
            else {
                // console.warn(">>>", JSON.stringify(token.token), JSON.stringify(value))
                // value = token.token.trim() // Trim surrounding whitespaces
                value = value.trim();
                var numVal = Number(value);
                if (constants_1.SEPARATORS.indexOf(value) >= 0 || value === constants_1.TILDE) {
                    type = 'sep';
                }
                else if (value === constants_1.DATASEP) {
                    type = 'datasep';
                }
                // When validating isNaN, check value is not a blank ''.
                // When the value is blank, Number(value) will set the numVal to 0
                else if (!isNaN(numVal) && value.trim() !== '') {
                    value = numVal;
                    type = 'number';
                }
                else if (value === constants_1.TRUE || value === constants_1.TRUE_F) {
                    value = true;
                    type = 'boolean';
                }
                else if (value === constants_1.FALSE || value === constants_1.FALSE_F) {
                    value = false;
                    type = 'boolean';
                }
                else if (value === constants_1.NULL || value === constants_1.NULL_F) {
                    value = null;
                    type = 'null';
                }
            }
            token.value = value;
            token.type = type;
            _this._tokens.push(token);
            _this.skipNextWhiteSpaces();
            return token;
        };
        this._next = function () {
            // Advance the step
            _this._col += 1;
            var index = ++_this._index;
            // Return token when text ends
            var ch = _this._text[index];
            var nextCh = _this._text[index + 1];
            // Identify char types
            var isWS = ch <= constants_1.SPACE; // Is white space or control char
            var isNewLine = ch === constants_1.NEW_LINE; // Is new line
            var isNextNewLine = nextCh === constants_1.NEW_LINE;
            var isSep = constants_1.SEPARATORS.indexOf(ch) >= 0; // Is separator
            var isNextSep = constants_1.SEPARATORS.indexOf(nextCh) >= 0;
            var isCollectionSep = ch === constants_1.TILDE;
            var isNextCollectionSep = nextCh === constants_1.TILDE;
            var isStarted = _this._start !== -1;
            var isDataSep = _this._text.substr(index - 2, 3) === constants_1.DATASEP;
            var isNextDataSep = _this._text.substr(index + 1, 3) === constants_1.DATASEP;
            if (is_1.isUndefined(ch)) {
                // Throw and error when escaping is not closed on last char
                if (_this._isEscaping) {
                    throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.incompleteEscapeSequence, 'End of the text reached before finishing the escape sequence.');
                }
                _this._done = true;
                return _this._returnToken();
            }
            // console.log("###", this._isCommenting, ch)
            // While the comment mode is active!
            if (_this._isCommenting) {
                _this._end = index;
                _this._value += ch;
                // Comment mode ends with new line, hence, turn it off when
                // a new line char is encountered.
                if (isNextNewLine) {
                    _this._isCommenting = false;
                    return _this._returnToken();
                }
                _this._tokenLength += 1;
                return _this._next();
            }
            var token = _this._lastToken;
            if (!token)
                return null; // Bypass TS check
            // Handle white-spaces
            if (isWS) {
                // Normalize newline to \n char when other newline modes such as '\r\n' & '\r' found
                if (ch === '\r') {
                    // Ignore current \r when a newline found in \r\n pair
                    if (nextCh === '\n') {
                        return _this._next();
                    }
                    ch = '\n'; // Else replace \r wit= '\n' // Else replace \r with \n
                    isNewLine = true;
                }
                if (_this._tokenLength > 0)
                    _this._value += ch;
                // Update values in case of new line
                if (isNewLine) {
                    _this._row += 1;
                    _this._col = 0;
                }
                if (!_this._isQuoatedString && !_this._isRawString && isStarted) {
                    if (isNextSep || isNextCollectionSep || isNextDataSep) {
                        if (_this._col === 0)
                            _this._col = 1;
                        return _this._returnToken();
                    }
                    else if (nextCh === constants_1.HASH) {
                        // this._isCommenting = true
                        return _this._returnToken();
                    }
                }
                return _this._next();
            }
            // // See if whitespace is escaped! Such as " \n  \t \n "
            // if (this._start === -1 && this._isRawString === false) {
            //   if(ch === BACKSLASH) {
            //   }
            // }
            // If not whitespace
            // =================
            _this._end = index;
            // Processing not started yet!
            if (_this._start === -1) {
                _this._start = index;
                // Set the row and col if not set yet.
                if (!token.col) {
                    token.index = _this._start;
                    token.col = _this._col;
                    token.row = _this._row;
                }
            }
            _this._tokenLength += 1;
            // Process the first char
            if (_this._tokenLength === 1) {
                // Start raw string when @" is found
                // if (ch === AT && nextCh === STRING_ENCLOSER) {
                //   this._isRawString = true
                //   return this._next()
                // }
                // Start a raw string when ' is found
                if (ch === constants_1.QUOTE) {
                    _this._isRawString = true;
                    return _this._next();
                }
                // When the " is encountered at first char,
                // activate enclosed string mode
                else if (ch === constants_1.STRING_ENCLOSER) {
                    _this._isQuoatedString = true;
                    return _this._next();
                }
            }
            // Handle Regular Strings
            if (_this._isQuoatedString) {
                if (ch === constants_1.STRING_ENCLOSER) {
                    // Start the string
                    // if (this._tokenLength === 1) {
                    //   this._isQuoatedString = true
                    //   return this._next()
                    // }
                    // End the string
                    if (_this._isEscaping === false) {
                        _this._isQuoatedString = false;
                        return _this._returnToken();
                    }
                }
                // Handle string escapes when not a raw string
                if (ch === constants_1.BACKSLASH && _this._isEscaping === false) {
                    _this._isEscaping = true;
                    return _this._next();
                }
                // When escaping, escape next char!
                if (_this._isEscaping) {
                    _this._isEscaping = false;
                    _this._value += ch in escapeCharMap ? escapeCharMap[ch] : ch;
                    return _this._next();
                }
            }
            // Handle Raw Strings
            if (_this._isRawString) {
                if (ch === constants_1.QUOTE) {
                    // Start the string
                    // if (this._tokenLength === 1) {
                    //   this._isRawString = true
                    //   return this._next()
                    // }
                    // When the first of two consicutive '' are found in the raw string,
                    // activate escaping mode and skip to the next char
                    if (_this._isEscaping === false && nextCh === constants_1.QUOTE) {
                        _this._isEscaping = true;
                        return _this._next();
                    }
                    // When the second of '' are found during escaping mode, just add the signle `
                    if (_this._isEscaping) {
                        _this._isEscaping = false;
                        _this._value += "'";
                        return _this._next();
                    }
                    // End the string
                    if (_this._isEscaping === false) {
                        _this._isRawString = false;
                        return _this._returnToken();
                    }
                }
                _this._value += ch;
                return _this._next();
            }
            // Character processing during open mode!
            if (!_this._isQuoatedString && !_this._isRawString) {
                // Initiate the commenting mode when
                if (nextCh === constants_1.HASH) {
                    _this._isCommenting = true;
                    return _this._returnToken();
                }
                if (ch === constants_1.HASH) {
                    _this._isCommenting = true;
                    _this._value += ch;
                    return _this._next();
                }
                if (isSep || isNextSep || isCollectionSep || isNextCollectionSep) {
                    _this._value += ch;
                    return _this._returnToken();
                }
                // Check for data separator
                else if (ch === constants_1.HYPHEN) {
                    // let value = this._text.substring(this._start, this._end + 1)
                    if (isDataSep) {
                        _this._value = '---';
                        return _this._returnToken();
                    }
                }
            }
            _this._value += ch;
            return _this._next();
        };
        this._text = text;
    }
    Object.defineProperty(Tokenizer.prototype, "done", {
        get: function () {
            return this._done;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Tokenizer.prototype, "length", {
        get: function () {
            return this._tokens.length;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(Tokenizer.prototype, "tokens", {
        get: function () {
            return this._tokens;
        },
        enumerable: false,
        configurable: true
    });
    Tokenizer.prototype.get = function (index) {
        return this._tokens[index];
    };
    Tokenizer.prototype.skipNextWhiteSpaces = function () {
        var ch = this._text[this._index + 1];
        // console.log("wow", ch, this._text.length, this._index)
        // Return token when text ends
        if (is_1.isUndefined(ch)) {
            this._done = true;
            return;
        }
        // If next ch is not ws, return
        if (ch > constants_1.SPACE)
            return;
        // Advance the step
        this._col += 1;
        ++this._index;
        if (this._tokenLength > 0)
            this._value += ch;
        // Update values in case of new line
        if (ch === constants_1.NEW_LINE) {
            this._row += 1;
            this._col = 0;
        }
        this.skipNextWhiteSpaces();
    };
    Object.defineProperty(Tokenizer.prototype, "_node", {
        get: function () {
            return {
                index: this._index,
                row: this._row,
                col: this._col
            };
        },
        enumerable: false,
        configurable: true
    });
    return Tokenizer;
}());
exports.default = Tokenizer;
//# sourceMappingURL=tokenizer.js.map