"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var io_error_1 = require("../errors/io-error");
var io_error_codes_1 = require("../errors/io-error-codes");
var is_1 = require("../utils/is");
var utils_1 = require("./utils");
var NUMBER_TYPES = ['number', 'int', 'int32', 'int16', 'byte'];
// age?: { number, true, 10, min:10, max:20}
/**
 * Represents the various number related data types
 * (such as number, int, byte, int16, int32) in Internet Object.
 *
 * @internal
 */
var NumberDef = /** @class */ (function () {
    function NumberDef(type) {
        var _this = this;
        if (type === void 0) { type = 'number'; }
        this.serialize = function (data, memberDef) {
            if (NUMBER_TYPES.indexOf(memberDef.type) === -1) {
                throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidType);
            }
            return _this.validate(data, memberDef).toString();
        };
        this._type = type;
        this._validator = _getValidator(type);
    }
    NumberDef.prototype.getType = function () {
        return this._type;
    };
    NumberDef.prototype.parse = function (data, memberDef, vars) {
        return this.validate(data, memberDef, vars);
    };
    NumberDef.prototype.load = function (data, memberDef) {
        return this.validate(data, memberDef);
    };
    NumberDef.prototype.validate = function (data, memberDef, vars) {
        var node = is_1.isToken(data) ? data : undefined;
        var value = node ? node.value : data;
        return _validate(this._validator, memberDef, value, node, vars);
    };
    return NumberDef;
}());
// Performs following validations.
//  * - Value is number, int, int32, int16, byte
//  * - Value is optional
//  * - Value is nullable
//  * - Value >= schema.min
//  * - Value <= schema.max
//  * - Value is in choices
function _validate(validator, memberDef, value, node, vars) {
    if (vars && is_1.isString(value)) {
        var valueFound = vars.getV(value);
        value = valueFound !== undefined ? valueFound : value;
    }
    var validatedData = utils_1.doCommonTypeCheck(memberDef, value, node);
    if (validatedData !== value || validatedData === undefined)
        return validatedData;
    if (!is_1.isNumber(value)) {
        throw new io_error_1.InternetObjectValidationError(io_error_codes_1.default.invalidValue);
    }
    else if (isNaN(Number(value))) {
        throw new io_error_1.InternetObjectValidationError(io_error_codes_1.default.notANumber, "Invalid number encountered for " + memberDef.path);
    }
    validator(memberDef, value, node);
    if (memberDef.min !== undefined) {
        var min = memberDef.min;
        if (memberDef.min !== undefined && value < min) {
            throw new (io_error_1.InternetObjectValidationError.bind.apply(io_error_1.InternetObjectValidationError, __spreadArrays([void 0], _invlalidMin(memberDef, value, node))))();
        }
    }
    if (memberDef.max !== undefined && value > memberDef.max) {
        throw new (io_error_1.InternetObjectValidationError.bind.apply(io_error_1.InternetObjectValidationError, __spreadArrays([void 0], _invlalidMax(memberDef, value, node))))();
    }
    return value;
}
function _intValidator(min, max, memberDef, value, node) {
    // Validate for integer
    if (value % 1 !== 0) {
        throw new (io_error_1.InternetObjectValidationError.bind.apply(io_error_1.InternetObjectValidationError, __spreadArrays([void 0], _notAnInt(memberDef, value, node))))();
    }
    if ((min !== -1 && value < min) || (max !== -1 && value > max)) {
        throw new (io_error_1.InternetObjectValidationError.bind.apply(io_error_1.InternetObjectValidationError, __spreadArrays([void 0], _outOfRange(memberDef, value, node))))();
    }
}
function _getValidator(type) {
    switch (type) {
        case 'number': {
            return function (memberDef, value, node) {
                if (!is_1.isNumber(value) && isNaN(Number(value))) {
                    throw new (io_error_1.InternetObjectValidationError.bind.apply(io_error_1.InternetObjectValidationError, __spreadArrays([void 0], _notANumber(memberDef, value, node))))();
                }
            };
        }
        case 'byte': {
            var range = Math.pow(2, 8); // -128,127
            return _intValidator.bind(null, -(range / 2), range / 2 - 1);
        }
        case 'int16': {
            var range = Math.pow(2, 16); // -32768,32767
            return _intValidator.bind(null, -(range / 2), range / 2 - 1);
        }
        case 'int32': {
            var range = Math.pow(2, 32); // -2147483648,2147483647
            return _intValidator.bind(null, -(range / 2), range / 2 - 1);
        }
        case 'int': {
            // Any non fraction number!
            return _intValidator.bind(null, -1, -1);
        }
        default: {
            console.assert(false, 'Invalid number type!');
        }
    }
}
function _outOfRange(memberDef, value, node) {
    return [
        io_error_codes_1.default.outOfRange,
        "The value (" + value + ") set for \"" + memberDef.path + "\" is out of range.",
        node
    ];
}
function _notAnInt(memberDef, value, node) {
    return [
        io_error_codes_1.default.notAnInteger,
        "Expecting an integer value for \"" + memberDef.path + "\", Currently it is " + value + ".",
        node
    ];
}
function _notANumber(memberDef, value, node) {
    return [
        io_error_codes_1.default.notANumber,
        "Expecting a number value for \"" + memberDef.path + "\", Currently it is " + value + ".",
        node
    ];
}
function _invlalidMin(memberDef, value, node) {
    return [
        io_error_codes_1.default.invalidMinValue,
        "The \"" + memberDef.path + "\" must be greater than or equal to " + memberDef.min + ", Currently it is " + value + ".",
        node
    ];
}
function _invlalidMax(memberDef, value, node) {
    return [
        io_error_codes_1.default.invalidMaxValue,
        "The \"" + memberDef.path + "\" must be less than or equal to " + memberDef.max + ", Currently it is " + value + ".",
        node
    ];
}
exports.default = NumberDef;
//# sourceMappingURL=number.js.map