"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var io_error_codes_1 = require("../errors/io-error-codes");
var typedef_registry_1 = require("./typedef-registry");
var utils_1 = require("./utils");
var is_1 = require("../utils/is");
var io_error_1 = require("../errors/io-error");
/**
 * Represents the ObjectTypeDef which is reponsible for parsing,
 * validating, loading and serializing Objects.
 */
var ObjectDef = /** @class */ (function () {
    function ObjectDef() {
        var _this = this;
        this._keys = null;
        /**
         * Parses the object in IO format into JavaScript object.
         */
        this.parse = function (data, memberDef, vars) {
            var value = is_1.isParserTree(data) ? data.values : undefined;
            return _this._process(memberDef, value, data, vars);
        };
        /**
         * Loads the JavaScript object.
         */
        this.load = function (data, memberDef) {
            return _this._process(memberDef, data);
        };
        /**
         * Serializes the object into IO format.
         */
        this.serialize = function (data, memberDef, isRoot) {
            if (isRoot === void 0) { isRoot = false; }
            if (memberDef.type !== 'object') {
                throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidObject);
            }
            var validatedData = utils_1.doCommonTypeCheck(memberDef, data);
            var serialized = [];
            var schema = memberDef.schema;
            schema.keys.forEach(function (key, index) {
                var memberDef = schema.defs[key];
                var typeDef = typedef_registry_1.TypedefRegistry.get(memberDef.type);
                var value = validatedData[key];
                var serializedValue = typeDef.serialize(value, memberDef);
                serialized.push(serializedValue);
            });
            if (isRoot) {
                // join array and trim off last commas, from the end
                return serialized.join(',').replace(/,+$/g, '');
            }
            // join array and trim off last commas, from the end
            return "{" + serialized.join(',').replace(/,+$/g, '') + "}";
        };
        // Process the parse and load requests
        this._process = function (memberDef, value, node, vars) {
            var validatedData = utils_1.doCommonTypeCheck(memberDef, value, node);
            if (validatedData !== value || validatedData === null || validatedData === undefined) {
                return validatedData;
            }
            var schema = memberDef.schema;
            var object = {};
            var fn = is_1.isParserTree(node) ? 'parse' : 'load';
            // When indexMode is on, members are read/loaded from the index.
            var indexMode = true;
            if (is_1.isParserTree(node)) {
                node.values.forEach(function (dataItem, index) {
                    var key;
                    var memberDef;
                    var dataValue;
                    if (is_1.isKeyVal(dataItem)) {
                        indexMode = false;
                        key = dataItem.key;
                        memberDef = schema.defs[key];
                        dataValue = dataItem.value;
                    }
                    // Process members only when the indexMode is true.
                    else if (indexMode || dataItem === undefined) {
                        key = schema.keys[index];
                        memberDef = schema.defs[key];
                        dataValue = dataItem;
                    }
                    else {
                        throw new io_error_1.InternetObjectSyntaxError(io_error_codes_1.default.positionalMemberAfterKeywordMember);
                    }
                    // When memberDef is not found, ignore such member
                    if (is_1.isUndefined(memberDef))
                        return;
                    var typeDef = typedef_registry_1.TypedefRegistry.get(memberDef.type);
                    object[key] = typeDef.parse(dataValue, memberDef, vars);
                });
                // Process the members who have not been included in
                // the data item.
                schema.keys.forEach(function (key) {
                    if (key in object)
                        return;
                    var memberDef = schema.defs[key];
                    var typeDef = typedef_registry_1.TypedefRegistry.get(memberDef.type);
                    object[key] = typeDef.parse(undefined, memberDef, vars);
                });
            }
            else {
                var keys = schema.keys;
                keys.forEach(function (key) {
                    var memberDef = schema.defs[key];
                    // When memberDef is not found, ignore such member
                    if (is_1.isUndefined(memberDef))
                        return;
                    var typeDef = typedef_registry_1.TypedefRegistry.get(memberDef.type);
                    var dataItem = value[key];
                    object[key] = typeDef.load(dataItem, memberDef);
                });
            }
            return object;
        };
    }
    /**
     * Returns the type this instance is going to handle.
     * Always returns object
     */
    ObjectDef.prototype.getType = function () {
        return 'object';
    };
    return ObjectDef;
}());
exports.default = ObjectDef;
//# sourceMappingURL=object.js.map