"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var io_error_1 = require("../errors/io-error");
var is_1 = require("../utils/is");
var utils_1 = require("./utils");
var io_error_codes_1 = require("../errors/io-error-codes");
// Reference: RFC 5322 Official Standard
// http://emailregex.com
var emailExp = /(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\])/;
// http://urlregex.com
var urlExp = /((([A-Za-z]{3,9}:(?:\/\/)?)(?:[\-;:&=\+\$,\w]+@)?[A-Za-z0-9\.\-]+|(?:www\.|[\-;:&=\+\$,\w]+@)[A-Za-z0-9\.\-]+)((?:\/[\+~%\/\.\w\-_]*)?\??(?:[\-\+=&;%@\.\w_]*)#?(?:[\.\!\/\\\w]*))?)/;
/**
 * Represents the StringTypeDef which is reponsible for parsing,
 * validating, loading and serializing strings.
 *
 * It performs the following validation
 * - Value is string
 * - Value is optional
 * - Value is nullable
 * - Value length <= maxLength
 * - Value length >= minLength
 * - Value is in choices
 */
var StringDef = /** @class */ (function () {
    function StringDef(type) {
        if (type === void 0) { type = 'string'; }
        this._type = type;
    }
    /**
     * Returns the type this instance is going to handle.
     * The return value could be any of the "string", "email", or "url"
     */
    StringDef.prototype.getType = function () {
        return this._type;
    };
    /**
     * Parses the string in IO format into JavaScript strings.
     */
    StringDef.prototype.parse = function (data, memberDef, vars) {
        return this.validate(data, memberDef, vars);
    };
    /**
     * Loads the JavaScript string.
     */
    StringDef.prototype.load = function (data, memberDef) {
        return this.validate(data, memberDef);
    };
    /**
     * Serializes the string into IO format.
     */
    StringDef.prototype.serialize = function (data, memberDef) {
        var value = this.validate(data, memberDef);
        var regexSep = /[~,:\{\}\[\]\@]|(?:---)/g;
        // When a separator is found in the data,
        // enclose the result string in a quoration
        if (regexSep.exec(value) !== null) {
            value = '"' + value + '"';
        }
        return value;
    };
    StringDef.prototype.validate = function (data, memberDef, vars) {
        var node = is_1.isToken(data) ? data : undefined;
        var value = node ? node.value : data;
        return _process(memberDef, value, node, vars);
    };
    return StringDef;
}());
exports.default = StringDef;
function _process(memberDef, value, node, vars) {
    // Replace vars
    if (vars) {
        var valueFound = vars.getV(value);
        value = valueFound || value;
    }
    // Run common check
    var validatedData = utils_1.doCommonTypeCheck(memberDef, value, node);
    if (validatedData !== value || validatedData === null || validatedData === undefined) {
        return validatedData;
    }
    // Validate
    if (!is_1.isString(value)) {
        throw new io_error_1.InternetObjectError(io_error_codes_1.default.notAString);
    }
    // TODO: Validate Data for subtypes
    _validatePattern(memberDef, value, node);
    var maxLength = memberDef.maxLength;
    // Max length check
    if (maxLength !== undefined && is_1.isNumber(maxLength)) {
        if (value.length > maxLength) {
            throw new io_error_1.InternetObjectValidationError(io_error_codes_1.default.invalidMaxLength, "Invalid maxLength for " + memberDef.path + ".");
        }
    }
    var minLength = memberDef.minLength;
    // Max length check
    if (minLength !== undefined && is_1.isNumber(minLength)) {
        if (value.length > minLength) {
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidMinLength, "Invalid minLength for " + memberDef.path + ".");
        }
    }
    return value;
}
function _validatePattern(memberDef, value, node) {
    var type = memberDef.type;
    // Validate user defined pattern
    if (type === 'string' && memberDef.pattern !== undefined) {
        var re = memberDef.re;
        if (!re) {
            var pattern = memberDef.pattern;
            // Add ^ and $ at the begging and end of the pattern respectively
            // if these characters are not set in the pattern
            pattern = pattern.startsWith('^') ? pattern : "^" + pattern;
            pattern = pattern.endsWith('$') ? pattern : pattern + "$";
            // Compile the expression and cache it into the memberDef
            re = memberDef.re = new RegExp(pattern);
        }
        if (!re.test(value)) {
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidValue);
        }
    }
    // Validate email
    else if (type === 'email') {
        if (!emailExp.test(value)) {
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidEmail);
        }
    }
    // Validate url
    else if (type === 'url') {
        if (!urlExp.test(value)) {
            throw new io_error_1.InternetObjectError(io_error_codes_1.default.invalidUrl);
        }
    }
}
function _notAString(path, data) {
    return [io_error_codes_1.default.notAString, "Expecting a string value for \"" + path + "\"", data];
}
function _invlalidChoice(path, data, choices) {
    return [
        io_error_codes_1.default.invalidValue,
        "The value of \"" + path + "\" must be one of the [" + choices.join(',') + "]. Currently it is " + data.value + ".",
        data
    ];
}
function _invlalidMinLength(path, data, minLength) {
    return [
        io_error_codes_1.default.invalidMinLength,
        "The length of \"" + path + "\" must be " + minLength + " or more. Currently it is " + data.value.length + ".",
        data
    ];
}
function _invlalidMaxLength(path, data, maxLength) {
    return [
        io_error_codes_1.default.invalidMaxLength,
        "The length of \"" + path + "\" must be " + maxLength + " or more. Currently it is " + data.value.length + ".",
        data
    ];
}
//# sourceMappingURL=string.js.map